/*********************************************************************
 * pMCimg
 *
 * Parallelized version of tMCimg.
 *
 *********************************************************************/

/*
 * This file is part of tMCimg.
 * 
 * tMCimg is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

#include "config.h"
#include "tMCimg.h"
#include "mccore.h"

#if (HAVE_ASSERT_H)
#include <assert.h>
#endif

#if (HAVE_UNISTD_H)
#include <unistd.h>
#endif

#if (! HAVE_PERROR)
/* Easier than endless #ifdef's */
# define perror(x) fprintf(stderr, "%s\n", x)
#endif

#ifndef USE_MPI
# error "MPI not requested, but trying to build parallel code"
#else
/* Get defines specific to the parallel code */
# include "parallel.h"
#endif

#define USAGE printf("optionss: if string contains...\n"\
          "\t? - display the option help (this)\n"\
          "\tn - inhibit .2pt file storage and dump\n"\
          "\tr - Use normal C ordering for segmentation/.2pt files\n"\
          "\ts - don't move source to the boundary\n"\
          "\tt - write exact exit time to .his.t instead of tgate # to .his\n"\
          "\tm - make tissue-air boundaries perfectly mirrored\n\n")

/*********************************************************************
 Function prototypes
 *********************************************************************/

static void run_simulation(const struct Config *, const TISSUE ***,
                           TWOPT_T *, TWOPT_T *, TWOPT_T *, TWOPT_T *, 
                           const char *);

extern void run_photon(const struct Config *, int, const TISSUE ***,
		       TWOPT_T *, TWOPT_T *, FILE *, int *);

void write_photon(const struct Config *, FILE *, double, double, double,
		  double, double, double, double, float *);

extern char *mpibuff;

/*********************************************************************
 BEGIN PROGRAM CODE
 *********************************************************************/

int main(int argc, char *argv[])
{
  struct Config syscfg;	      /* System configuration, constructed
			       *   from the .cfg file */
  TWOPT_T *II, *JJ;	      /* FOR STORING THE 2-PT FLUENCE */
  TWOPT_T *II0,*JJ0;	      /* Extra copy for master program */
  TISSUE ***tissueType;	      /* STORE THE IMAGE FILE */
  int i, j, n, nvox;

  /* INITIALIZE THE CONFIGURATION STRUCTURE'S MEMORY */

  initialize_config(&syscfg);

  /* SET UP THE FLOATING POINT CO-PROCESSOR */

  fpusetup();

  /* SET UP THE MPI LIBRARY */

  initParallel(&argc, &argv);	/* Call before parsing arguments! */

#if 0
  if (getNodeCount() <= 2)
    {
      fprintf(stderr, 
              "Parallelized code does not support nodecount %d <= 2\n",
	      getNodeCount());
      fprintf(stderr, "Call tMCimg for single-threaded simulations instead\n");
      MPI_Abort(MPI_COMM_WORLD, 1);
    }
#endif

  /* GET THE COMMAND LINE ARGUMENT(S) */

  if (argc == 1)
    {
      printf("tMCimg version 8.0%c\n", 'A' + 0);
      printf("USAGE:: tMCimg [-opts] input_file (.cfg assumed)\n");

      USAGE;

      printf("Default options:\n");
      printf("\t2pt files %s be generated\n",
	     (syscfg.flags.gen_twopt != 0) ? "will" : "will not");
      printf("\t%s-ordered data files\n",
	     (syscfg.flags.matlab_order != 0) ? "Matlab" : "C");
      printf("\tSources %s to interface\n",
	     (syscfg.flags.source_move != 0) ? "moved" : "not moved");
      printf("\tDetectors %s to interface\n",
	     (syscfg.flags.detector_move != 0) ? "moved" : "not moved");
      printf("\tExiting photons %s to interface\n",
	     (syscfg.flags.exiting_move != 0) ? "moved" : "not moved");
      printf("\t%s recorded in history file\n",
	     (syscfg.flags.exact_exit_time != 0) ?
	     "Exact exit times" : "Time indices");
      if (syscfg.flags.mirror != 0)
	printf("\tMirrored air boundaries\n");
      printf("\n");

      return 1;
    }
  else
    for (i = 1; i < argc; i++)
      {
	if (argv[i][0] == '-')	  /* introduces option flags */
	  {
	    /* search through command line option words to find
	     * meaningful chars... */

	    for (j = 1; argv[i][j] != '\0'; j++)
	      switch (argv[i][j])
		{
		  case 'm':
		    syscfg.flags.mirror = !syscfg.flags.mirror;
		    printf("Tissue-air boundaries %s be mirrored.\n",
			   (syscfg.flags.mirror ? "will" : "will not"));
		    break;
		  case 'n':
		    syscfg.flags.gen_twopt = !syscfg.flags.gen_twopt;
		    printf("The .2pt file %s be generated.\n",
			   (syscfg.flags.gen_twopt ? "will" : "will not"));
		    break;
		  case 'r':
		    syscfg.flags.matlab_order = !syscfg.flags.matlab_order;
		    printf("Data %s be saved in matlab order.\n",
			   (syscfg.flags.matlab_order ? "will" : "will not"));
		    break;
		  case 'd':
		    syscfg.flags.detector_move = !syscfg.flags.detector_move;
		    printf("Detectors %s be moved to interface.\n",
			   (syscfg.flags.
			    detector_move ? "will" : "will not"));
		    break;
		  case 's':
		    syscfg.flags.source_move = !syscfg.flags.source_move;
		    printf("Source %s be moved to interface.\n",
			   (syscfg.flags.source_move ? "will" : "will not"));
		    break;
		  case 'e':
		    syscfg.flags.exiting_move = !syscfg.flags.exiting_move;
		    printf("Exiting photons %s be moved to boundary.\n",
			   (syscfg.flags.exiting_move ? "will" : "will not"));
		    break;
		  case 't':
		    syscfg.flags.exact_exit_time =
		      !syscfg.flags.exact_exit_time;
		    printf("History file %s record exact exit times.\n",
			   (syscfg.flags.
			    exact_exit_time ? "will" : "will not"));
		    break;
		  case 'h':
		  case '?':
		    USAGE;
		    return 0;
		  default:
		    printf("Unknown option -%c ignored\n", argv[i][j]);
		}
	  }
	else
	  {
	    /* Assume that since it isn't a command-line flag that it's
	     *  the name of the configuration file.  Diddle the argument
	     *  vector to fool the parser into doing the right thing.
	     */

	    argv[1] = argv[i];
	    argv[2] = NULL;	  /* may or may not be necessary */

	    argc = 2;

	    break;
	  }
      }

  /* PARSE THE INPUT FILE */

  if ((n = loadconfig(&syscfg, argv[1])) != 0)
    MPI_Abort(MPI_COMM_WORLD, n);

  if (syscfg.NT < 1)
    {
      fprintf(stderr, "Error: total number of photons must be >1 [is %d]\n",
	syscfg.NT);
      exit(1);
    }

  rescale_system(&syscfg);

  /* SUMMARY OF EXTANT OPTIONS */

  if (syscfg.flags.gen_twopt)
    printf("...2pt files will be generated\n");
  if (syscfg.flags.matlab_order)
    printf("...Matlab-ordered data files\n");
  if (syscfg.flags.source_move)
    printf("...Sources moved to interface\n");
  if (syscfg.flags.detector_move)
    printf("...Detectors moved to interface\n");
  if (syscfg.flags.exiting_move)
    printf("...Exiting photons moved to interface\n");
  if (syscfg.flags.exact_exit_time)
    printf("...Exact exit times recorded in history file\n");
  if (syscfg.flags.mirror)
    printf("...Mirrored air boundaries\n");

  /* READ IN THE SEGMENTED DATA FILE */

  if ((n = loadsegments(&syscfg, &tissueType)) != 0)
    {
      MPI_Abort(MPI_COMM_WORLD, n);
      exit(n);			  /* probably not reached */
    }

  /* ALLOCATE SPACE FOR AND INITIALIZE THE PHOTON FLUENCE TO 0 */

  nvox = syscfg.nIxstep * syscfg.nIystep * syscfg.nIzstep * syscfg.nTstep;
  II = II0 = NULL;
  JJ = JJ0 = NULL;

  if (syscfg.flags.gen_twopt)
    {
      if ((II  = alloc_1d_array(nvox,sizeof(TWOPT_T))) == NULL)
	MPI_Abort(MPI_COMM_WORLD, 1);

      if ((II0 = alloc_1d_array(nvox,sizeof(TWOPT_T))) == NULL)
	MPI_Abort(MPI_COMM_WORLD, 1);

      /* CW/TD don't need this.  Hopefully, the viewer etc. will be
       *  smart enought to know the difference */

      if (syscfg.frequency != 0.0)
	{
	  if ((JJ  = alloc_1d_array(nvox,sizeof(TWOPT_T))) == NULL)
	    MPI_Abort(MPI_COMM_WORLD, 1);
	  if ((JJ0 = alloc_1d_array(nvox,sizeof(TWOPT_T))) == NULL)
	    MPI_Abort(MPI_COMM_WORLD, 1);
	}

      /* INITIALIZE THE FLUENCE DATA */

      if (II != NULL)
	for (i = 0; i < nvox; i++)
	  II[i] = 0.0;

      if (JJ != NULL)
	for (i = 0; i < nvox; i++)
	  JJ[i] = 0.0;

      if (II != NULL && II0 != NULL)
	memcpy(II0, II, nvox*sizeof(TWOPT_T));

      if (JJ != NULL && JJ0 != NULL)
	memcpy(JJ0, JJ, nvox*sizeof(TWOPT_T));
    }

  /* MAKE SURE THE SOURCES ARE AT INTERFACES */

  if (syscfg.flags.source_move)
    move_sources(&syscfg, (const TISSUE ***)tissueType);

  /* MAKE SURE THE DETECTORS ARE AT INTERFACES */

  if (syscfg.flags.detector_move)
    move_detectors(&syscfg, (const TISSUE ***)tissueType);

  /* RUN THE MONTE CARLO SIMULATION */

  run_simulation(&syscfg, (const TISSUE ***)tissueType, 
				II, JJ, II0, JJ0, argv[1]);

  /* DONE */

  endParallel();

  nvox = syscfg.nIxstep * syscfg.nIystep * syscfg.nIzstep * syscfg.nTstep;

  free_1d_array(II,  nvox);
  free_1d_array(JJ,  nvox);
  free_1d_array(II0, nvox);
  free_1d_array(JJ0, nvox);

  return 0;
}

/* ------------------------ RUN THE SIMULATION ---------------------- */

static void run_simulation(const struct Config *syscfg, 
                           const TISSUE ***tissueType,
                           TWOPT_T *II, TWOPT_T *JJ, 
                           TWOPT_T *II0, TWOPT_T *JJ0, const char *basefile)
{
  FILE *fp = NULL;		  /* FILE POINTERS FOR SAVING THE DATA */
  int i, ns, nvox, seed;
  unsigned long N;
#if (REALLY_RANDOM)
  FILE *rp;
#endif
  unsigned long NPhotons;
  int n;

#if (REALLY_RANDOM)
  /* /dev/random is a much better seed for parallel code than seed+n */

  if ((rp = fopen("/dev/random", "rb")) == NULL)
    {
      if ((rp = fopen("/dev/urandom", "rb")) == NULL)
	{
	  fread(&seed, sizeof(seed), 1, rp);
	  fclose(rp);
	}
      else
	{
	  printf("Warning: can't open /dev/random, using less random seed\n");
	  seed += myNode();
	}
    }
  else
    {
      fread(&seed, sizeof(seed), 1, rp);
      fclose(rp);
    }
#else
  seed = syscfg->seed;
#endif

  nvox = syscfg->nIystep * syscfg->nIxstep * syscfg->nIzstep * syscfg->nTstep;

  if (syscfg->nSrcs > 1)
    {
      if (syscfg->nDets == 0)
	printf("No detectors: writing all exiting photons to %s.his.aN\n",
	       basefile);
      else if (syscfg->flags.exact_exit_time)
	printf("Writing exact exit times to %s.his.tN\n", basefile);
      else
	printf("Writing exiting tgate number to %s.his.N\n", basefile);
    }
  else
    {
      if (syscfg->nDets == 0)
	printf("No detectors: writing all exiting photons to %s.his.a\n",
	       basefile);
      else if (syscfg->flags.exact_exit_time)
	printf("Writing exact exit times to %s.his.t\n", basefile);
      else
	printf("Writing exiting tgate number to %s.his\n", basefile);
    }

  /* The fraction of the total that I'm responsible for.  The number of
   * nodes is so vastly smaller than the number of photons, we won't
   * worry about rounding problems.  Subtract 1 since the master
   * thread doesn't itself run any photons */

  if (getNodeCount() > 1)
    NPhotons = syscfg->NT / (getNodeCount() - 1);
  else
    NPhotons = syscfg->NT;

#if (! REALLY_RANDOM)
  /* Diddle my random number seed, or there's no point to running
   * in parallel! */

  seed += myNode();
#endif

  /* FOR EVERY SOURCE, RUN THE SIMULATION */

  for (ns = 0; ns < syscfg->nSrcs; ns++)
    {
      /*******************************************************************
       * START MIGRATING THE PHOTON GENERATING PHOTONS UNTIL NUMBER OF
       * PHOTONS EXECUTED (N) IS EQUAL TO THE NUMBER TO BE GENERATED (NT)
       *******************************************************************/

      MPI_Barrier(MPI_COMM_WORLD);       /* Start all the nodes together */

      if (isMaster())
	{
          int nleft = 0;

	  if ((fp = openhis(ns, basefile, syscfg)) == NULL)
	    {
	      fprintf(stderr, "Can't open history file %s\n", basefile);
	      MPI_Abort(MPI_COMM_WORLD, 1);
	    }

	  /* As photons leave the system at the different nodes, write
	   * them to the history file.  Only the master copy should
	   * write to the history file to preserve data integrity. */

	  /* nleft does not include myself */
	  nleft = getNodeCount() - 1;

	  while (nleft > 0)
	    {
	      MPI_Status status;
	      float buffer[MAX_HISTENT];

	      /* Gather in photons as they leave the system
	       * and write history to disk (blocking call) */

	      MPI_Recv(buffer, MAX_HISTENT, MPI_FLOAT, MPI_ANY_SOURCE,
		       TAG_HISFILE, MPI_COMM_WORLD, &status);

	      MPI_Get_count(&status, MPI_FLOAT, &n);

	      /* We use empty records to signify that a node
	       *  has finished it's allocation of photons */

	      if (n == 0)
		{
#if (DEBUG)
		  printf("Node %d finished\n", status.MPI_SOURCE);
#endif
		  nleft -= 1;
		}
	      else
		{
		  int nw;

#if (PARANOID)
                  if (syscfg->nDets == 0)
                    assert(n == 4+syscfg->Ntissue);
                  else
                    assert(n == 2+syscfg->Ntissue);
#endif

		  if ((nw = fwrite(buffer, sizeof(float), n, fp)) != n)
		    {
		      perror("Error writing to history file");
		      MPI_Abort(MPI_COMM_WORLD, 1);
		    }
		}
	    }

          /* CLOSE HISTORY FILE */
          fclose(fp);
	}
      else			  /* not master process */
	{
	  float foo;		  /* Dummy memory for MPI_Send */

          for (N = 0; N < NPhotons; N++)
            run_photon(syscfg, ns, tissueType, II, JJ, fp, &seed);

	  /* Signal that I have no more photons to deliver */
	  MPI_Send(&foo, 0, MPI_FLOAT, 0, TAG_HISFILE, MPI_COMM_WORLD);

          /* Send any buffered messages and deallocate the buffer */

	  if (mpibuff != NULL)
	    {
              extern int mpiblen;
              void *p;
              int   s;

	      /* Detach buffer, blocks until all messages have been sent */
	      MPI_Buffer_detach(&p, &s);

	      free_1d_array(mpibuff, 100 * mpiblen);

	      mpibuff = NULL;
              mpiblen = 0;
	    }
	}

      /* SAVE THE PHOTON DENSITY (2-POINT GREENS FUNCTION) */

      if (syscfg->flags.gen_twopt)
	{
	  /* Sum all the nodes into the master's copy */

	  if (II != NULL)
	    {
#if (PARANOID)
	      assert(II  != 0);
	      assert(II0 != 0);
#endif
	      MPI_Reduce(II, II0, 
                         nvox, MPI_TWOPT, MPI_SUM, 0, MPI_COMM_WORLD);
	    }

	  if (JJ != NULL)
	    {
#if (PARANOID)
	      assert(JJ  != 0);
	      assert(JJ0 != 0);
#endif
	      MPI_Reduce(JJ, JJ0, 
                         nvox, MPI_TWOPT, MPI_SUM, 0, MPI_COMM_WORLD);
	    }

	  if (II != NULL && II0 != NULL)
	    memcpy(II, II0, nvox * sizeof(TWOPT_T));

	  if (JJ != NULL && JJ0 != NULL)
	    memcpy(JJ, JJ0, nvox * sizeof(TWOPT_T));

	  /* Save the final (summed) version of the 2pt file */

	  if (isMaster())
	    {
	      if (syscfg->nSrcs == 1)
		save2pt(basefile, -1, II, JJ, nvox, NULL);
	      else
		save2pt(basefile, ns, II, JJ, nvox, NULL);
	    }

	  /* Reset memory for next pass through */

          if (II != NULL)
            {
              for (i = 0; i < nvox; i++)
                II[i] = 0.0;

              if (II0 != NULL)
                memcpy(II0, II, nvox * sizeof(TWOPT_T));
            }

          if (JJ != NULL)
            {
              for (i = 0; i < nvox; i++)
                JJ[i] = 0.0;

              if (JJ0 != NULL)
                memcpy(JJ0, JJ, nvox * sizeof(TWOPT_T));
            }
	}
    }

  return;
}

/* Call the parallel savehis() routine */

void write_photon(const struct Config *syscfg, FILE *fp,
		  double x, double y, double z,
		  double kx, double ky, double kz, 
		  double Ttot, float *lenTiss)
{
#if (PARANOID)
  assert(!isMaster());
#endif

  parallel_savehis(syscfg, x, y, z, kx, ky, kz, Ttot, lenTiss);

  return;
}

